#
# Socket Implementation
#

newSOCKnode <- function(machine = "localhost", ...,
                        options = defaultClusterOptions) {
    # **** allow some form of spec here
    # **** make sure options are quoted
    options <- addClusterOptions(options, ...)
    port <- getClusterOption("port", options)
    rhome <- getClusterOption("rhome", options)
    script <- file.path(getClusterOption("scriptdir", options), "RSOCKnode.sh")
    rshcmd <- getClusterOption("rshcmd", options)
    user <- getClusterOption("user", options)
    env <- paste("RPROG=", getClusterOption("rprog", options),
                 " MASTER=", getClusterOption("master", options),
                 " PORT=", port,
                 " OUT=", getClusterOption("outfile", options),
                 " R_LIBS=", getClusterOption("rlibs", options),
                 sep="")
    system(paste(rshcmd, "-l", user, machine, "env", env, script))
    con <- socketConnection(port = port, server=TRUE, blocking=TRUE,
                            open="a+b")
    structure(list(con = con, host = machine), class = "SOCKnode")
}

makeSOCKmaster <- function() {
    master <- Sys.getenv("MASTER")
    port <- as.integer(Sys.getenv("PORT"))
    # maybe use `try' and slep/retry if first time fails?
    con <- socketConnection(master, port = port, blocking=TRUE, open="a+b")
    structure(list(con = con), class = "SOCKnode")
}

closeNode.SOCKnode <- function(node) close(node$con)

sendTag.SOCKnode <- function(node, tag) write(tag, node$con)
recvTag.SOCKnode <- function(node) readLines(node$con, 1)
sendData.SOCKnode <- function(node, data) serialize(data, node$con)
recvData.SOCKnode <- function(node) unserialize(node$con)

makeSOCKcluster <- function(names, ..., options = defaultClusterOptions) {
    if (! exists("serialize") && ! require(serialize))
        stop("the `serialize' package is needed for SOCK clusters.")
    options <- addClusterOptions(options, ...)
    cl <- vector("list",length(names))
    for (i in seq(along=cl))
        cl[[i]] <- newSOCKnode(names[[i]], options = options)
    cl
}
